package org.jutils.collection;
import java.util.*;
import org.jutils.collection.validators.*;

/**
 * This Map implementor combines the functionality of Observable and the Validators API. The Observable part is
 * correspondable for handling and firing change events when values have been changed, and keys added and removed.<p>
 * Furthermore, validators for key and values may be applied to check the correctness of put keys and values (at put time only).
 * For an example of usage, see the main method.
 * @author M.Winkler, doubleslash NetBusiness GmbH, Friedrichshafen
 */
public class ExtendedMap implements Map {

   /** This is the inner map which actually holds the key-value pairs. */
   private java.util.Map map_ = null;
   private java.util.Collection keyValidators_ = null;
   private java.util.Collection valueValidators_ = null;

   private java.util.Collection observers_ = null;





/**
 * Creates a new EktendedMap with a default Map class.
 */
protected ExtendedMap() {
   map_ = new HashMap();

}





/**
 * Generates a new ExtendedMap, providing some config information.
 * @author M.Winkler, doubleslash NetBusiness GmbH, Friedrichshafen
 * @param innerMapClass java.lang.Class The Class of the map type to be used internally.
 * @param keyValidators java.util.Collection A predefined set of key validators, if any. May be null.
 * @param valueValidators java.util.Collection A predefined set of value validators, if any. May be null.
 * @exception java.lang.IllegalAccessException An exception opccurred when instantiating the map class.
 * @exception java.lang.InstantiationException An exception opccurred when instantiating the map class.
 */
public ExtendedMap(
   Class innerMapClass,
   Collection keyValidators,
   Collection valueValidators)
   throws java.lang.IllegalAccessException, java.lang.InstantiationException {

   // check map class parameter
   if (innerMapClass != null)
     map_ = (Map) innerMapClass.newInstance();
   else
     map_ = new HashMap();

   // check the other parameters.
   if (keyValidators != null)
     keyValidators_ = new HashSet(keyValidators);
   if (valueValidators != null)
     valueValidators = new HashSet(valueValidators);

}





/**
 * Creates a new ExtendedMap with a default map class and the provided validators.
 * @author M.Winkler, doubleslash NetBusiness GmbH, Friedrichshafen
 * @param keyValidators java.util.Collection A set of key validators, if not null.
 * @param valueValidators java.util.Collection A set of value validators, if not null.
 */
public ExtendedMap(Collection keyValidators, Collection valueValidators) {
   map_ = new HashMap();
   if (keyValidators != null)
     keyValidators_ = new HashSet(keyValidators);
   if (valueValidators != null)
     valueValidators_ = new HashSet(valueValidators);
}





/**
 * Startet die Anwendung.
 * @param args ein Array von Befehlszeilenargumenten
 */
public static void main(java.lang.String[] args) {
   System.out.println("Testing ExtendedMap ...");
   int errorCount = 0;

   try {

     ExtendedMap map = new ExtendedMap();

     //  test validator integration
     map.addKeyValidator(new ClassValidator(Integer.class, false));
     map.addValueValidator(new ClassValidator(String.class, false));

     System.out.println("Test putting an invalid key into it...");
     try {
       map.put("blabla", "hoscheho");
       System.out.println("ERROR --- invalid key accepted !");
       errorCount++;
     } catch (IllegalArgumentException iae) {
       // thats what should happen
     }

     // Test putting an invalid value into it
     System.out.println("Test putting an invalid value into it...");
     try {
       map.put(new Integer(5), new Integer(5));
       System.out.println("ERROR --- invalid value accepted !");
       errorCount++;
     } catch (IllegalArgumentException iae) {
       // thats what should happen
     }

     // Test putting a valid pait into it
     System.out.println("Test putting a valid pair into it...");
     try {
       map.put(new Integer(2), "hoscheho");
       String value = (String) map.get(new Integer(2));
       if (!value.equals("hoscheho")) {
         System.out.println("ERROR --- valid pair accepted, but not stored in map!");
         errorCount++;
       }
     } catch (IllegalArgumentException iae) {
       System.out.println("ERROR --- valid pair not accepted! " + iae.toString());
       errorCount++;
     }

     // Test clearing the map
     System.out.println("Test clearing the map ...");
     try {
       map.clear();
       if (map.size() > 0) {
         System.out.println("ERROR --- size() returns a positive number!");
         errorCount++;
       }
     } catch (IllegalArgumentException iae) {
       System.out.println("ERROR --- valid pair not accepted! " + iae.toString());
       errorCount++;
     }

   } catch (RuntimeException rte) {
     System.out.println("ERROR --- uncaught exception :" + rte.toString());
     rte.printStackTrace();
     errorCount++;

   } finally {

     // stats
     System.out.println("----------------------------");
     System.out.println("" + errorCount + " errors detected.");
     System.out.println("----------------------------");
   }

}





/**
 * Adds a new validator to the list of key validators.
 * @author M.Winkler, doubleslash NetBusiness GmbH, Friedrichshafen
 * @param newValidator org.jutils.collection.validators.Validator A new key validator.
 * @see org.jutils.collection.validators.Validator
 */
public void addKeyValidator(Validator newValidator) {

   // check key validators attribute
   if (keyValidators_ == null)
     keyValidators_ = new HashSet();

   keyValidators_.add(newValidator);
}





/**
 * Adds a new Map observer to the observer list. Each observer can only be once in the list. The notification order is not defined.
 * @author M.Winkler, doubleslash NetBusiness GmbH, Friedrichshafen
 * @param newObserver org.jutils.collection.MapObserver
 * @see org.jutils.collection.MapObserver
 */
public void addMapObserver(MapObserver newObserver) {
   if (observers_ == null)
     observers_ = new HashSet();
   observers_.add(newObserver);
}





/**
 * Adds a new validator to the list of Value validators.
 * @author M.Winkler, doubleslash NetBusiness GmbH, Friedrichshafen
 * @param newValidator org.jutils.collection.validators.Validator A new value validator.
 * @see org.jutils.collection.validators.Validator
 */
public void addValueValidator(Validator newValidator) {

   // check key validators attribute
   if (valueValidators_ == null)
     valueValidators_ = new HashSet();

   valueValidators_.add(newValidator);
}





/**
 * @see java.util.Map
 */
public void clear() {
   map_.clear();
   pFireMapCleared();
}





/**
 * @see java.util.Map
 */
public boolean containsKey(java.lang.Object key) {
   return map_.containsKey(key);
}





/**
 * @see java.util.Map
 */
public boolean containsValue(java.lang.Object value) {
   return map_.containsValue(value);
}





/**
 * @see java.util.Map
 */
public java.util.Set entrySet() {
   return map_.entrySet();
}





/**
 * @see java.util.Map
 */
public java.lang.Object get(java.lang.Object key) {
   return map_.get(key);
}





/**
 * @see java.util.Map
 */
public boolean isEmpty() {
   return map_.isEmpty();
}





/**
 * @see java.util.Map
 */
public java.util.Set keySet() {
   return map_.keySet();
}





/**
 * Fires the "key added"-event to all observers.
 * @author M.Winkler, doubleslash NetBusiness GmbH, Friedrichshafen
 * @param key java.lang.Object A validated key that already has been added to the map.
 * @param value java.lang.Object The corresponding value
 */
private void pFireKeyAdded(Object key, Object value) {
   if (observers_ != null) {
     Iterator iterator = observers_.iterator();
     while (iterator.hasNext())
        ((MapObserver) iterator.next()).keyAdded(this, key, value);
   }
}





/**
 * Fires the "key removed"-event to all observers.
 * @author M.Winkler, doubleslash NetBusiness GmbH, Friedrichshafen
 * @param key java.lang.Object A validated key that has been removed from the map.
 * @param value java.lang.Object The corresponding value
 */
private void pFireKeyRemoved(Object key, Object value) {
   if (observers_ != null) {
     Iterator iterator = observers_.iterator();
     while (iterator.hasNext())
        ((MapObserver) iterator.next()).keyRemoved(this, key, value);
   }
}





/**
 * Fires the "key added"-event to all observers.
 * @author M.Winkler, doubleslash NetBusiness GmbH, Friedrichshafen
 * @param key java.lang.Object A validated key that already has been added to the map.
 * @param value java.lang.Object The corresponding value
 */
private void pFireMapCleared() {
   if (observers_ != null) {
     Iterator iterator = observers_.iterator();
     while (iterator.hasNext())
        ((MapObserver) iterator.next()).mapCleared(this);
   }
}





/**
 * Fires the "value changed "-event to all observers.
 * @author M.Winkler, doubleslash NetBusiness GmbH, Friedrichshafen
 * @param key java.lang.Object A validated key.
 * @param oldValue java.lang.Object The old value.
 * @param newValue java.lang.Object The new value.
 */
private void pFireValueChanged(Object key, Object oldValue, Object newValue) {
   if (observers_ != null) {
     Iterator iterator = observers_.iterator();
     while (iterator.hasNext())
        ((MapObserver) iterator.next()).valueChanged(this, key, oldValue, newValue);
   }
}





/**
 * @see java.util.Map
 */
public java.lang.Object put(java.lang.Object key, java.lang.Object value) {

   // validate key
   try {
     pValidateKey(key);
   } catch (InvalidTypeException ite) {
     throw new IllegalArgumentException("key '" + key + "' denied :" + ite);
   } catch (InvalidValueException ive) {
     throw new IllegalArgumentException("key '" + key + "' denied :" + ive);
   }

   // validate value
   try {
     pValidateValue(value);
   } catch (InvalidTypeException ite) {
     throw new IllegalArgumentException("value '" + value + "' denied :" + ite);
   } catch (InvalidValueException ive) {
     throw new IllegalArgumentException("value '" + value + "' denied :" + ive);
   }

   Object oldObject = map_.put(key, value);

   if (oldObject != null)
     pFireValueChanged(key, oldObject, value);
   else
     pFireKeyAdded(key, value);

   return oldObject;
}





/**
 * @deprecated Sorry, not yet implemented (will come soon)
 * @see java.util.Map
 */
public void putAll(java.util.Map t) {
   throw new UnsupportedOperationException("putAll not yet implemented.");

}





/**
 * Validates a single key to match all key validators.
 * @author M.Winkler, doubleslash NetBusiness GmbH, Friedrichshafen
 * @param keyTovalidate java.lang.Object A key that is validated if it applies to the restrictions that are set by the validators.
 * @exception org.jutils.collection.validators.InvalidValueException Thrown if the passed Key has an invalid value.
 * @exception org.jutils.collection.validators.InvalidTypeException Thrown if the passed Key has an invalid type.
 */
public void pValidateKey(Object keyTovalidate)
   throws
     org.jutils.collection.validators.InvalidValueException,
     org.jutils.collection.validators.InvalidTypeException {

   // check if we have validators
   if ((keyValidators_ != null) || (keyValidators_.size() > 0)) {
     Iterator iterator = keyValidators_.iterator();

     // since we throw out the exceptions, this one really is short
     while (iterator.hasNext())
        ((Validator) iterator.next()).checkValid(keyTovalidate);
   }
}





/**
 * Validates a single value to match all value  validators.
 * @author M.Winkler, doubleslash NetBusiness GmbH, Friedrichshafen
 * @param valueToValidate java.lang.Object A value that is validated if it applies to the restrictions that are set by the validators.
 * @exception org.jutils.collection.validators.InvalidValueException Thrown if the passed value has an invalid value.
 * @exception org.jutils.collection.validators.InvalidTypeException Thrown if the passed value has an invalid type.
 */
public void pValidateValue(Object valueToValidate)
   throws
     org.jutils.collection.validators.InvalidValueException,
     org.jutils.collection.validators.InvalidTypeException {

   // check if we have validators
   if ((valueValidators_ != null) || (valueValidators_.size() > 0)) {
     Iterator iterator = valueValidators_.iterator();

     // since we throw out the exceptions, this one really is short
     while (iterator.hasNext())
        ((Validator) iterator.next()).checkValid(valueToValidate);
   }
}





/**
 * @see java.util.Map
 */
public java.lang.Object remove(java.lang.Object key) {
   Object value = map_.remove(key);

   if (value != null)
     pFireKeyRemoved(key, value);

   return value;
}





/**
 * Removes a specific validator from the list of key validators.
 * @author M.Winkler, doubleslash NetBusiness GmbH, Friedrichshafen
 * @param validator org.jutils.collection.validators.Validator A key validator to remove.
 * @see org.jutils.collection.validators.Validator
 */
public void removeKeyValidator(Validator validator) {

   if (keyValidators_ != null)
     keyValidators_.remove(validator);
}





/**
 * removes a Map observer from the observer list.
 * @author M.Winkler, doubleslash NetBusiness GmbH, Friedrichshafen
 * @param observer org.jutils.collection.MapObserver
 * @see org.jutils.collection.MapObserver
 */
public void removeMapObserver(MapObserver observer) {
   if (observers_ != null)
     observers_.remove(observer);
}





/**
 * Removes a specific validator from the list of value validators.
 * @author M.Winkler, doubleslash NetBusiness GmbH, Friedrichshafen
 * @param validator org.jutils.collection.validators.Validator A value  validator to remove.
 * @see org.jutils.collection.validators.Validator
 */
public void removeValueValidator(Validator validator) {

   if (valueValidators_ != null)
     valueValidators_.remove(validator);
}





/**
 * @see java.util.Map
 */
public int size() {
   return map_.size();
}





/**
 * @see java.util.Map
 */
public java.util.Collection values() {
   return map_.values();
}
}
