package org.jutils.collection.validators;
import java.util.*;

/**
* This validator checks for values to be (or not to be) in a given set of elements.
* You can pass that set of (il)legal values with a constructor, and specify wether this is an inclusion or en exclusion list.
* If an exclusion list, then the value is considered valid if it is NOT in the given set.
* @author M.Winkler, doubleslash NetBusiness GmbH, Friedrichshafen
*/
public class SetValidator extends AbstractValidator {
   private Set elements_ = null;
   private boolean exclude_ = false;





/**
 * Creates a new SetValidator.
 * @author M.Winkler, doubleslash NetBusiness GmbH, Friedrichshafen
 * @param set java.util.Set A set of elements, with which the values are checked. The set is NOT cloned, so you can modify the contents afterwards.
 * @param exclude boolean If true, then a checked value only is true if its NOT in the set.
 */
public SetValidator(Set set, boolean exclude) {
   if (set == null)
     throw new IllegalArgumentException("passed set is null.");
   elements_ = set;
   exclude_ = exclude;
}





/**
 * Startet die Anwendung.
 * @param args ein Array von Befehlszeilenargumenten
 */
public static void main(java.lang.String[] args) {
   System.out.println("Checking SetValidator ...");
   Set set = new HashSet();
   SetValidator includeValidator = new SetValidator(set, false);
   int errorCount = 0;

   // check include validator
   // --------------------------------
   // check empty include validator
   System.out.println();
   System.out.println("First checking inclusion validator");
   System.out.println();
   System.out.println("Checking empty validator ...");
   try {
     includeValidator.checkValid("blabla");
     System.out.println("ERROR --- invalid value returned ok!");
     errorCount++;
   } catch (InvalidValueException ive) {
     // this is what should happen
   } catch (InvalidTypeException ite) {
     // this is what should NOT happen
     System.out.println("ERROR --- invalid value returned wrong exception!");
     errorCount++;
   }

   // check a single value
   set.add("blabla");
   System.out.println("Checking a single value...");
   try {
     includeValidator.checkValid("blabla");
   } catch (InvalidValueException ive) {
     // this is what should NOT happen
     System.out.println("ERROR --- valid value did fail!");
     errorCount++;
   } catch (InvalidTypeException ite) {
     // this is what should NOT happen
     System.out.println("ERROR --- valid value did returned 'wrong type'!");
     errorCount++;
   }

   // check null
   System.out.println("Checking null ...");
   try {
     includeValidator.checkValid(null);
     System.out.println("ERROR --- null value returned ok!");
     errorCount++;
   } catch (InvalidValueException ive) {
     // this is what should happen
   } catch (InvalidTypeException ite) {
     // this is what should NOT happen
     System.out.println("ERROR --- null value returned wrong exception!");
     errorCount++;
   }

   // stats
   System.out.println("--------------------------");
   System.out.println("" + errorCount + " errors detected.");
   System.out.println("--------------------------");
   System.exit(errorCount);
}





/**
 * @see org.jutils.collection.validators.Validator
 */
public void checkValid(java.lang.Object value)
   throws InvalidValueException, InvalidTypeException {

   // check if the element is in the list
   boolean valid = elements_.contains(value);

   // check if we are excluding
   if (exclude_)
     valid = !valid;

   // if we are invalid, throw exception.
   if (!valid)
     throw new InvalidValueException(
       exclude_ ? "element in set" : "element not in set");
}





/**
 * Code, der bei der Bereinigung dieses Objekts ausgefhrt wird.
 *
 * Jede von einer Beendigungsmethode bergebene Ausnahmebedingung stoppt die
 * Beendigung. Andernfalls wird er ignoriert.
 */
protected void finalize() throws Throwable {
   // Hier Code einfgen, um den Empfnger zu beenden.
   // Diese Implementierung leitet die Nachricht einfach nur an die Superkomponente weiter. Sie knnen dies ersetzen oder ergnzen.
   super.finalize();
}





/**
 * Gibt eine Zeichenfolge zurck, die den Wert dieses Objekts darstellt.
 * @return eine Zeichenfolgedarstellung des Empfngers
 */
public String toString() {

   StringBuffer returnValue = new StringBuffer(30);
   returnValue.append("SetValidator checking values for being ");
   if (exclude_)
     returnValue.append("excluded from ");
   else
     returnValue.append("included in ");

   returnValue.append(elements_.toString());

   return returnValue.toString();

}
}
