package org.jutils.runtime;
import java.text.*;
import java.util.*;
import org.jutils.collection.*;
import java.io.*;

/**
* This is a helper class to implement a recycling of the base classes of java.lang and java.util.
* Basically, this also could be implemented by using according org.jutils.pool.ResourceWrapper,
* but this one is to heavy weight to wrap Long, so no one would do this.
* @author: M.Winkler, doubleSlash GmbH
*/
public final class Recycler { /**
* The class name of the Vector, to avoid String instance generation. This one is private.
*/
   private static final String CN_VECTOR = Vector.class.getName();
   /**
   * The class name of the List interface, to avoid String instance generation. May be used to specify a recycler.
   */
   public static final String CN_LIST = List.class.getName();
   /** The initial capacity of the recycler instances. */
   public static final int INITIAL_RECYCLER_SIZE = 1000;
   /**
   * Contains some <tt>Long</tt> instances.
   */
   private static LongHashtable longPool_ = null;

   /** Defines wether or not the long pool should be used. */
   private static boolean longRecyclerEnabled_ = true;

   /** The number of calls to getLong. */
   private static int callCountGetLong_ = 0;

   /** The initial capacity of the long recycler. */
   public static final int INITIAL_LONG_RECYCLER_SIZE = 2500;

   /** Contains the upper border for the capacity of the long recycler. */
   private static int maxCountLongRecycler_ = INITIAL_LONG_RECYCLER_SIZE * 1000;
   /**
   * Contains some <tt>java.util.Vector</tt> instances.
   */
   private static List vectorRecycler_ = null;
   /**
   * Contains the generic object recycler. The types are identified by their package and class name.
   * The instances are stored in a java.util.List in a row. also stored here are the upper nspecific
   * recyclers, except the long recycler, which has another concept.
   */
   private static Map genericRecycler_ = null;



/**
 * Returns a <tt>Long</tt> instance containing the specified value, trying to get it from the recycler. <br>
 * You donnot have to "recycle" a Long instance, you only have to
 * make sure that this instance is coming from here. Otherwise you will get a multiple instances.
 * Creation date: (08.11.2000 09:03:03)
 * @return java.lang.Long An Instance of <tt>Long</tt>. Never null.
 * @param value long A value that should be wrapped by the returned Long instance.
 */
public static Long getLong(long value) {
   Long returnValue = null;
   callCountGetLong_++;
   if (longRecyclerEnabled_) {
     if (longPool_ == null) {

       // create a new long recycler
       setupLongRecycler(maxCountLongRecycler_, INITIAL_LONG_RECYCLER_SIZE);
     } else {

       // return without storing the value
       returnValue = (Long) longPool_.get(value);
     }
   }

   if (returnValue == null) {

     // create a new instance
     returnValue = new Long(value);

     // put it into the pool, if still free capacity available.
     if (longRecyclerEnabled_ && (longPool_.size() < maxCountLongRecycler_))
       longPool_.put(value, returnValue);
   }
   return returnValue;
}



   /**
    * Removes all elements from all caches. Does NOT remove recyclers from the generic recycler. Use destroyRecyclers ()instead.
    * @return The number of removed instances.
    * @see org.jutils.runtime.Recycler.destroyAllRecyclers()
    * @author: M.Winkler, doubleSlash GmbH
    */
   public static int clearAllRecyclers() {
      int removedCount = 0;

      // clear long recycler
      if (longPool_ != null) {
         removedCount += longPool_.size();
         longPool_.clear();
         callCountGetLong_ = 0;
      }

      // clear vector recycler
      if (vectorRecycler_ != null) {
         removedCount += vectorRecycler_.size();
         vectorRecycler_.clear();
      }

      // clear generic recycler
      if (genericRecycler_ != null) {

         // remove all elements
         Set keySet = genericRecycler_.keySet();
         Iterator keyIterator = keySet.iterator();
         while (keyIterator.hasNext()) {
            Object key = keyIterator.next();
            List instanceList = (List) genericRecycler_.remove(key);
            if (instanceList != null) {
               removedCount += instanceList.size();
               instanceList.clear();
            }
         }

      }
      return removedCount;
   }



   /**
    * Returns an instance of the passed type. Tries to get it from the recycler.If not available, then
    * tries to instantinate one.
    * @author: M.Winkler, doubleSlash GmbH
    * @return java.lang.Object
    * @param className java.lang.String
    * @exception IllegalArgumentException If the passed class name is an illegal value.
    */
   public static Object getInstance(String className)
      throws IllegalArgumentException {
      Object returnValue = null;

      // shortcut to vector
      if (className.equals(CN_VECTOR)) {
         returnValue = getVector();
      } else if (genericRecycler_ != null) {

         // wrap in a hash key wrapper.
         Object hashKey = CollectionHelper.hashKey(className);

         // look for the list instance in the generic recycler
         List instanceList = (List) genericRecycler_.get(hashKey);

         // chewck if there is something in it.
         if ((instanceList != null) && (instanceList.size() > 0))
            returnValue = instanceList.remove(instanceList.size() - 1);
      }

      // check wether we still have  nothing.
      if (returnValue == null) {

         // try to instantinate the class
         try {
            returnValue = Class.forName(className).newInstance();
         } catch (InstantiationException ie) {
            throw new IllegalArgumentException(
               "Cannot instantinate class : " + ie.toString());
         } catch (ClassNotFoundException cnfe) {
            throw new IllegalArgumentException(
               "Cannot instantinate class : " + cnfe.toString());
         } catch (IllegalAccessException iae) {
            throw new IllegalArgumentException(
               "Cannot instantinate class : " + iae.toString());
         }

      }
      return returnValue;
   }



   /**
    * Returns a <tt>Vector</tt> instance, but first tries to get it from the recycler.
    * @param int preferredCapacity A preferred capacity the vector should have. May be < 1 to surpress this.
    * @author: M.Winkler, doubleSlash GmbH
    * @return java.util.Vector
    */
   public static Vector getVector() {
      Vector returnValue = null;
      if ((vectorRecycler_ != null) && (vectorRecycler_.size() > 0)) {
         returnValue = (Vector) vectorRecycler_.remove(vectorRecycler_.size() - 1);
      } else if (returnValue == null)
         returnValue = new Vector();
      return returnValue;
   }



   /**
    * Prints the statistics to this printStream.
    * @author: M.Winkler, doubleSlash GmbH
    * @param printStream java.io.PrintStream
    */
   public static void listStatistics(PrintStream printStream) {
      printStream.println("Statistics of " + Recycler.class.getName());
      printStream.println("---------------------------------------");
      printStream.println((longPool_ != null ? "" + longPool_.size() : "no") + " elements in longPool_");
      printStream.println("" + callCountGetLong_ + " calls to getLong()");
   }



/**
 * Tests this class.
 * @param args ein Array von Befehlszeilenargumenten
 */
public static void main(java.lang.String[] args) {
   System.out.println("---------------------------------------------");
   System.out.println("Testing with Long pool disabled...");
   System.out.println("---------------------------------------------");
   Recycler.setLongRecyclerEnabled(false);
   int errorCount = test();

   System.out.println("---------------------------------------------");
   System.out.println("Testing with Long pool enabled ...");
   System.out.println("---------------------------------------------");
   Recycler.setLongRecyclerEnabled(true);
   errorCount += test();

   System.out.println("---------------------------------------------");
   System.out.println("" + errorCount + " errors detected totally.");
   System.out.println("---------------------------------------------");

}



   /**
    * @author: M.Winkler, doubleSlash GmbH
    */
   private static void pSetupRecyclers() {

      // create instance
      genericRecycler_ = new Hashtable(1000);

      // link in vector recycler
      if (vectorRecycler_ != null) {
         Object hashKey = CollectionHelper.hashKey(CN_VECTOR);
         genericRecycler_.put(hashKey, vectorRecycler_);
      }
   }



   /**
    * Puts this class into the generic recycler. Tests wether this is a vector, and therefore can be put into
    * the vector recycler.
    * @author: M.Winkler, doubleSlash GmbH
    * @param instance java.lang.Object
    */
   public static void recycle(Object instance) {
      recycle(instance, instance.getClass().getName());

   }



   /**
    * Puts this class into the generic recycler. Tests wether this is a vector, and therefore can be put into
    * the vector recycler.
    * @author: M.Winkler, doubleSlash GmbH
    * @param instance java.lang.Object The instance to recycle.
    * @param className java.lang.String The class name under which this instance should be put into the recycler.
    * Use this if you want to force a pooling of interfaces, which is not possible otherwise.
    */
   public static void recycle(Object instance, String className) {

      // shortcut to vector
      if (className.equals(CN_VECTOR)) {
         Vector v = (Vector) instance;
         recycle(v);
      } else {
         List instanceList = null;

         // for faster hashtable access, create a cached hash code
         Object hashKey = CollectionHelper.hashKey(className);

         // check for existing generic recyclers
         if (genericRecycler_ == null)
            pSetupRecyclers();
         else {
            instanceList = (List) genericRecycler_.get(hashKey);
         }

         // check wether we have to create an instance list
         if (instanceList == null) {

            // put the list into the generic recycler
            instanceList = getVector();
            genericRecycler_.put(hashKey, instanceList);
         }

         // put in the instance
         instanceList.add(instance);
      }
   }



   /**
    * Recycles a <tt>Vector</tt>, i.e. clears it, and puts it to the vector store.
    * @author: M.Winkler, doubleSlash GmbH
    * @param vector java.util.Vector
    */
   public static void recycle(Vector vector) {
      if (vector == null)
         throw new IllegalArgumentException("Parameter 'vector' is null.");
      vector.clear();
      if (vectorRecycler_ == null)
         vectorRecycler_ = new Vector(INITIAL_RECYCLER_SIZE);
      vectorRecycler_.add(vector);
   }



/**
 * Makes sure that the passed instance is no more in the recycler. You normally dont have to call this.
 * @author: M.Winkler, doubleSlash GmbH
 * @return boolean True if the element was present and has been removed, false if the element was NOT present.
 * @param instance java.lang.Long
 */
public static boolean remove(Long instance) {
   boolean returnValue = false;
   if (instance == null)
     throw new IllegalArgumentException("Parameter 'instance' is null.");
   if (longPool_ != null) {

     Long storedValue = (Long) longPool_.get(instance.longValue());

     // since we only have to remove the
     // instance if exactly the same instance (not only the wrapped value)
     // is present, we have to compare references.
     if (storedValue == instance) {
       longPool_.remove(instance.longValue());
       returnValue = true;
     }
   }
   return returnValue;
}



   /**
    * Sets new edge values for the long recycler.
    * @author: M.Winkler, doubleSlash GmbH
    * @param maxCount int
    * @param ensuredCapacity int A capacity that will be ensured after this call.
    */
   public static void setupLongRecycler(int maxCount, int ensuredCapacity) {
      if ((longPool_ != null) && (longPool_.size() > maxCount))
         throw new IllegalArgumentException("cannot set max count to a new value- recycler already contains more values.");
      if (maxCount < 1)
         throw new IllegalArgumentException("parameter 'maxCount' < 1");
      if (ensuredCapacity < 1)
         throw new IllegalArgumentException("parameter 'ensuredCapacity' is < 1");
      maxCountLongRecycler_ = maxCount;

      if (longPool_ != null)
         throw new IllegalStateException("The long pool already is initialized!");
      longPool_ = new LongHashtable(Math.min(ensuredCapacity, maxCount));
   }





/**
 * @author M.Winkler, doubleslash NetBusiness GmbH, Friedrichshafen
 * @return boolean
 */
public static boolean isLongRecyclerEnabled() {
   return longRecyclerEnabled_;
}





/**
 * @author M.Winkler, doubleslash NetBusiness GmbH, Friedrichshafen
 * @param newLongRecyclerEnabled boolean
 */
public static void setLongRecyclerEnabled(boolean newLongRecyclerEnabled) {
   longRecyclerEnabled_ = newLongRecyclerEnabled;
}





/**
 * Tests this class.
 * @param args ein Array von Befehlszeilenargumenten
 */
private static int test() {
   System.out.println("Testing Recycler ...");
   int errorCount = 0;
   try {

     int testCount = 50000;
     // Recycler.setupLongRecycler(testCount * 2, testCount * 2);

     org.jutils.time.StopWatch stop = new org.jutils.time.StopWatch();

     System.out.println("Testing getLong()");

     for (int index = 0; index < testCount; index++) {
       Long aLong = getLong(index);
       if (aLong.longValue() != index) {
         System.out.println("!!! first access : Value mismatch!");
         errorCount++;
       }

     }

     System.out.println("Time needed to create " + testCount + " instances:" + stop.nice());

     for (int index = 0; index < testCount; index++) {
       Long aLong = getLong(index);
       if (aLong.longValue() != index) {
         System.out.println("!!! second access : Value mismatch!");
         errorCount++;
       }
     }

     System.out.println("Time needed to get " + testCount + " instances:" + stop.nice());
     Recycler.listStatistics(System.out);

     for (int index = 0; index < testCount; index++) {
       Long aLong = new Long(index);
       if (aLong.longValue() != index) {
         System.out.println("!!! second access : Value mismatch!");
         errorCount++;
       }
     }
     System.out.println("Time needed to create " + testCount + ":" + stop.nice());
   } catch (RuntimeException rte) {
     System.out.println("Uncaught Exception : " + rte);
     errorCount++;
   }
   return errorCount;
}
}