package org.jutils.runtime;

/**
 * Insert the type's description here.
 * Creation date: (09.10.2000 10:35:44)
 * @author:
 */
import java.lang.*;
import java.util.*;
public class ThreadPool {
    private static ThreadPool instance_ = null;
    public static final int LOW_PRIORITY = 0;
    public static final int NORMAL_PRIORITY = 1;
    public static final int HIGH_PRIORITY = 2;
    private List[] idleThreads_;
    private boolean shutDown_ = false;
    private int threadCreationCounter_;
    private boolean debug_ = false;



    /**
         * @author: M.Winkler, doubleSlash NetBusiness GmbH
         *
         */
        private ThreadPool() {

            // generating idleThreads.
            List[] idleThreads = {new Vector(5), new Vector(5), new Vector(5)};
            idleThreads_ = idleThreads;
            threadCreationCounter_ = 0;
        }



    /**
         * Returns the number of newly created threads so far.
         * @author: M.Winkler, doubleSlash NetBusiness GmbH
         *
         * @return int
         */
        public int getCreatedThreadsCount() {
            return threadCreationCounter_;
        }



    /**
         * @author: M.Winkler, doubleSlash NetBusiness GmbH
         *
         * @return org.jutils.runtime.ThreadPool
         */
        public static ThreadPool instance() {
            if (instance_ == null)
                instance_ = new ThreadPool();
            return instance_;
        }



    /**
         * @author: M.Winkler, doubleSlash NetBusiness GmbH
         *
         * @return boolean
         */
        public boolean isDebug() {
            return debug_;
        }



    /**
     * @author: M.Winkler, doubleSlash NetBusiness GmbH
     *
     * @param args java.lang.String[]
     */
    public static void main(String[] args) {
        System.out.println("Testing ThreadPool ...");
        System.out.println("Creating ThreadPool ...");
        ThreadPool pool = ThreadPool.instance();
        pool.setDebug(true);
        class TestRunner implements Runnable {
            public int count = 0;
            public void run() {
                System.out.println("Testrunner sleeping 5 seconds ...");
                synchronized (this) {
                    try {
                        wait(5000);
                    }
                    catch (InterruptedException ioe) {
                    }
                }
                System.out.println("Testrunner leaving  ...");
                count++;
            }
        }
        System.out.println("Starting a new thread ...");
        TestRunner runner = new TestRunner();
        pool.start(runner, pool.HIGH_PRIORITY);
        System.out.println("count : " + runner.count);
        System.out.println("Thread count : " + pool.getCreatedThreadsCount());
        pool.shutdown();
    }



    /**
         * @author: M.Winkler, doubleSlash NetBusiness GmbH
         *
         * @param repoolingThread org.jutils.runtime.PooledThread
         */
        protected synchronized void repool(PooledThread repoolingThread) {
            if (!shutDown_) {
                if (debug_) {
                    System.out.println("ThreadPool.repool() : repooling ...");
                }
                switch (repoolingThread.getPriority()) {
                    case Thread.MIN_PRIORITY :
                        {
                            idleThreads_[LOW_PRIORITY].add(repoolingThread);
                            break;
                        }
                    case Thread.NORM_PRIORITY :
                        {
                            idleThreads_[NORMAL_PRIORITY].add(repoolingThread);
                            break;
                        }
                    case Thread.MAX_PRIORITY :
                        {
                            idleThreads_[HIGH_PRIORITY].add(repoolingThread);
                            break;
                        }
                    default :
                        throw new IllegalStateException("Illegal priority found while repooling a Thread!");
                }
                notifyAll();
            }
            else {
                if (debug_) {
                    System.out.println("ThreadPool.repool() : Destroying incoming thread....");
                }
                repoolingThread.shutDown();
            }
            if (debug_) {
                System.out.println("ThreadPool.recycle() : done.");
            }
        }



    /**
         * @author: M.Winkler, doubleSlash NetBusiness GmbH
         *
         * @param newDebug boolean
         */
        public void setDebug(boolean newDebug) {
            debug_ = newDebug;
        }



    /**
         * @author: M.Winkler, doubleSlash NetBusiness GmbH
         *
         */
        public synchronized void shutdown() {
            shutDown_ = true;
            if (debug_) {
                System.out.println("ThreadPool : shutting down ...");
            }
            for (int prioIndex = 0; prioIndex <= HIGH_PRIORITY; prioIndex++) {
                List prioThreads = idleThreads_[prioIndex];
                for (int threadIndex = 0; threadIndex < prioThreads.size(); threadIndex++) {
                    PooledThread idleThread = (PooledThread) prioThreads.get(threadIndex);
                    idleThread.shutDown();
                }
            }
            notifyAll();
            if (debug_) {
                System.out.println("ThreadPool : shutdown done.");
            }
        }



    /**
         * Searches for an idle thread, and passes the passed
         * target over to that PooledThread.
         * @author: M.Winkler, doubleSlash NetBusiness GmbH
         *
         * @param target java.lang.Runnable The runnable that should run once.
         * @param priority int
         */
        public synchronized void start(Runnable target, int priority) {
            PooledThread thread = null;
            List idleList = idleThreads_[priority];
            if (idleList.size() > 0) {

                // get it (and remove it) from idle list.
                int lastIndex = idleList.size() - 1;
                thread = (PooledThread) idleList.get(lastIndex);
                idleList.remove(lastIndex);
                thread.setTarget(target);
            }
            else {
                threadCreationCounter_++;
                thread = new PooledThread(target, "PooledThread #" + threadCreationCounter_, this);
                switch (priority) {
                    case LOW_PRIORITY :
                        {
                            thread.setPriority(Thread.MIN_PRIORITY);
                            break;
                        }
                    case NORMAL_PRIORITY :
                        {
                            thread.setPriority(Thread.NORM_PRIORITY);
                            break;
                        }
                    case HIGH_PRIORITY :
                        {
                            thread.setPriority(Thread.MAX_PRIORITY);
                            break;
                        }
                    default :
                        {
                            thread.setPriority(Thread.NORM_PRIORITY);
                            break;
                        }
                }
                thread.start();
            }
        }
}

