package org.jutils.text;
import java.util.*;
import org.jutils.collection.*;

/**
 * Implements an associator of string - string pairs.
 * The special thing about this is that you can search
 * for substring, and return the counterpart associated with
 * that substring.
 * @author: M.Winkler, doubleSlash GmbH
 */
public class FragmentMatcher {

   // The minimal length of the partial key
   public static final int DEFAULT_FRAGMENT_SIZE = 3;
   private long maxId = 0;
   private int fragmentSize_ = DEFAULT_FRAGMENT_SIZE;

   // contains the complete key-value-pairs
   List pairs_ = new LinkedList();

   // Contains the single three letter fragments
   private Associator fragments_ = null;

   /** The default mutex of the class, to avoid synchronizing on the
   * object, which can lead to deadlocks.
   **/
   private final java.lang.Object mutex_ = new java.lang.Object();

   /** Degfines wether space chars are treated as word separator, or as normal
   * characters. */
   private boolean ignoreSpaces_;

   /**
   * Defines wether or not this fragment matcher is case sensitive or not.
   */
   private boolean caseSensitive_ = true;

   /**
   * Defines wether or not already some elements have been added.
   */
   private boolean elementsAdded_ = false;



/**
      * Creates a new pattern matcher with a default fragment size.
      */
public FragmentMatcher() {
   super();
   ignoreSpaces_ = false;

   try {
     fragments_ = new Associator(TreeSet.class, false);
   } catch (Exception e) {
     System.out.println("Exception while intstantiating the fragment matcher:" + e.toString());
     e.printStackTrace(System.out);
     fragments_ = new Associator();
   }
}



/**
      * Creates a new pattern matcher with the defined fragment size, which must be > 1!
      */
public FragmentMatcher(int fragmentSize) {
   this();
   if (fragmentSize < 2)
     throw new IllegalArgumentException("fragmentSize MUST be greater 1!");
   fragmentSize_ = fragmentSize;
   ignoreSpaces_ = false;
}



/**
   * @author Michael Winkler, doubleSlash GmbH
   *
   * @param args java.lang.String[]
   */
public static void main(String[] args) {
   System.out.println("Test FragmentMatcher ...");
   int errorCount = 0;
   int fragmentSize = 5;

   // --------------------------------------------------------------
   long startTicks = System.currentTimeMillis();
   FragmentMatcher matcher = new FragmentMatcher(fragmentSize);
   System.out.println("test empty Fragment Matcher...");
   try {
     matcher.get("blubber");
   } catch (Exception e) {
     System.out.println("!!! empty Fragment Matcherdoes throw an exception : ");
     e.printStackTrace();
     errorCount++;

   }

   System.out.println("--- test empty Fragment Matcherwith empty value ...");
   try {
     matcher.get("");
   } catch (Exception e) {
     System.out.println("!!! empty Fragment Matcher does throw an exception : ");
     e.printStackTrace();
     errorCount++;

   }

   System.out.println("--- test putting some strings into it ...");
   String[] testArray = { "hahaha", "nununu", "tatata", "hanuta", "hatanu", "nutaha", "nuhata", "tanuha", "tahanu" };
   for (int index = 0; index < testArray.length; index++) {
     matcher.put(testArray[index], testArray[index]);
   }

   // --------------------------------------------------------------
   System.out.println("--- test searching substrings ...");
   for (int index = 0; index < testArray.length; index++) {
     String wholeString = testArray[index];
     String subString = wholeString;

     // "normal" test
     if (wholeString.length() >= matcher.getFragmentSize())
       subString = wholeString.substring(1, fragmentSize + 1);
     Set results = matcher.get(subString);
     if (!results.contains(wholeString)) {
       System.out.println("!!! Content mismatch in normal sub string test !");
       System.out.println("!!! whole/sub were " + wholeString + "/" + subString);
       System.out.println("!!! Results are :" + results);
       errorCount++;
     }
     Iterator resultsIterator = results.iterator();
     while (resultsIterator.hasNext()) {
       Object o = resultsIterator.next();
       String singleResult = (String) o;
       if (!(singleResult.indexOf(subString) > -1)) {
         System.out.println("!!! wrong result!");
         errorCount++;
       }
     }

     // extreme test
     subString = wholeString.substring(0, matcher.getFragmentSize());
     results = matcher.get(subString);
     if (!results.contains(wholeString)) {
       System.out.println("!!! Content mismatch in extreme test for word beginnings!");
       System.out.println("!!! whole/sub were " + wholeString + "/" + subString);
       System.out.println("!!! Results are :" + results);
       errorCount++;
     }

     resultsIterator = results.iterator();
     while (resultsIterator.hasNext()) {
       String singleResult = (String) resultsIterator.next();
       if (!(singleResult.indexOf(subString) > -1)) {
         System.out.println("!!! wrong result!");
         errorCount++;
       }
     }

     // multipler test
     subString = wholeString.substring(0, 2);
     subString = subString + subString;
     subString = subString + subString;
     results = matcher.get(subString);
     if (results.size() > 0) {
       if (!results.contains(wholeString)) {
         System.out.println("!!! test for multiplied strings did return values!");
         System.out.println("!!! whole/sub were " + wholeString + "/" + subString);
         System.out.println("!!! Results are :" + results);
         errorCount++;
       }
     }
   }
   String[] fragments = matcher.getFragmentsArray();
   Arrays.sort(fragments, String.CASE_INSENSITIVE_ORDER);
   StringBuffer outputLine = CollectionHelper.toStringBuffer(fragments, "{", ",", "}", null);
   System.out.println("--- fragments are :" + outputLine.toString());

   // Testing for ignoreSpaces
   matcher = new FragmentMatcher(2, true);
   Long first = new Long(1);
   Long second = new Long(2);
   matcher.put("You bloody son of a bitch!", first);
   matcher.put("Irene hat was an der Backe", second);

   // check for first word
   System.out.println("--- Checking FragmentMatcher with ignoring spaces ...");
   System.out.println("--- Checking for first word ('You')...");
   Set results = matcher.get("You");
   if (results.size() != 1) {
     errorCount++;
     System.out.println(
       "!!! Fragment Matcher(ignoring spaces) did not return correct count! Count is "
         + results.size()
         + ", but should be 1.");
   }
   if (!results.contains(first)) {
     errorCount++;
     System.out.println("!!! Fragment Matcher(ignoring spaces) did not return value! results are " + results);

   }

   // check for first word with spaces
   System.out.println("--- Checking for first word with whitespace ('You ') ...");
   results = matcher.get("You ");
   if (results.size() != 1) {
     errorCount++;
     System.out.println(
       "!!! When asking for first word and a space, Fragment Matcher(ignoring spaces) did not return correct count! Count is "
         + results.size()
         + ", but should be 1.");
   }
   if (!results.contains(first)) {
     errorCount++;
     System.out.println(
       "!!! When asking for first word and a space, Fragment Matcher(ignoring spaces) did not return value! results are "
         + results);

   }

   // check for last word
   System.out.println("--- Checking for last word ('bitch!')...");
   results = matcher.get("bitch!");
   if (results.size() != 1) {
     errorCount++;
     System.out.println(
       "!!! Fragment Matcherdid not return correct count! Count is " + results.size() + ", but should be 1.");
   }
   if (!results.contains(first)) {
     errorCount++;
     System.out.println("!!! Fragment Matcherdid not return value! results are " + results);

   }

   // check for last word with spaces
   System.out.println("--- Checking for last word with a leading whitespace (' bitch!')...");

   results = matcher.get(" bitch!");
   if (results.size() != 1) {
     errorCount++;
     System.out.println(
       "!!! Fragment Matcherdid not return correct count! Count is " + results.size() + ", but should be 1.");
   }
   if (!results.contains(first)) {
     errorCount++;
     System.out.println("!!! Fragment Matcher(ignoring spaces) did not return value! results are " + results);

   }

   // check for middle word without spaces
   System.out.println("--- Check for middle word without spaces ('son')...");
   results = matcher.get("son");
   if (results.size() != 1) {
     errorCount++;
     System.out.println(
       "!!! Fragment Matcherdid not return correct count! Count is " + results.size() + ", but should be 1.");
   }
   if (!results.contains(first)) {
     errorCount++;
     System.out.println("!!! Fragment Matcher(ignoring spaces) did not return value! results are " + results);

   }

   // check for middle word with spaces
   System.out.println("--- Check for middle word containing spaces (' son ')...");
   results = matcher.get(" son ");
   if (results.size() != 1) {
     errorCount++;
     System.out.println(
       "!!! Fragment Matcherdid not return correct count! Count is " + results.size() + ", but should be 1.");
   }
   if (!results.contains(first)) {
     errorCount++;
     System.out.println("!!! Fragment Matcher(ignoring spaces) did not return value! results are " + results);

   }

   // check for middle word with spaces
   System.out.println("--- Check for middle word containing spaces ('of a')...");
   results = matcher.get("of a");
   if (results.size() != 1) {
     errorCount++;
     System.out.println(
       "!!! Fragment Matcherdid not return correct count! Count is " + results.size() + ", but should be 1.");
   }
   if (!results.contains(first)) {
     errorCount++;
     System.out.println("!!! Fragment Matcher(ignoring spaces) did not return value! results are " + results);

   }

   // check for middle word with spaces, and surrounding spaces
   System.out.println("--- Check for middle word containing spaces (' of a ')...");
   results = matcher.get(" of a ");
   if (results.size() != 1) {
     errorCount++;
     System.out.println(
       "!!! Fragment Matcherdid not return correct count! Count is " + results.size() + ", but should be 1.");
   }
   if (!results.contains(first)) {
     errorCount++;
     System.out.println("!!! Fragment Matcher(ignoring spaces) did not return value! results are " + results);

   }
   // check for partial double words with spaces between
   System.out.println("--- check for partial double words with spaces between ('oody so')...");
   results = matcher.get("oody so");
   if (results.size() != 1) {
     errorCount++;
     System.out.println(
       "!!! Fragment Matcherdid not return correct count! Count is " + results.size() + ", but should be 1.");
   }
   if (!results.contains(first)) {
     errorCount++;
     System.out.println("!!! Fragment Matcher(ignoring spaces) did not return value! results are " + results);

   }

   // check for case sensitivity
   System.out.println("--- check for case sensitivity ...");

   results = matcher.get("OODY");
   if (results.size() > 0) {
     errorCount++;
     System.out.println("!!! Case sensitive Fragment Matcherdid return value! results are " + results);
   }

   // check for IllegalStateException when switching to case insensitivity
   System.out.println("--- check for IllegalStateException when switching to case insensitivity ...");
   try {
     matcher.setCaseSensitive(false);
     System.out.println("!!! IllegalStateException NOT thrown!");
     errorCount++;

   } catch (IllegalStateException ise) {
     // all fine.
   }

   // setup new fragment matcher and check for case insensitivity.
   System.out.println("--- check for case INsensitivity ...");
   matcher = new FragmentMatcher(4, true, false);

   matcher.put("abcd", new Object());
   results = matcher.get("ABcD");
   if ((results == null) || (results.size() < 1)) {
     System.out.println("!!! Case insensitive Fragment Matcher did return any results!");
     errorCount++;
   }

   // check mass fill speed
   org.jutils.time.StopWatch stop = new org.jutils.time.StopWatch();
   int count = 10000;
   System.out.println("--- check mass fill " + count + " items speed ...");
   matcher = new FragmentMatcher(2);
   for (int i = 0; i < count; i++) {
     String anyString = "00000" + i;
     matcher.put(anyString, anyString);
   }

   System.out.println("Needed " + stop.nice() + " to fill in " + count + " pieces.");

   long stopTicks = System.currentTimeMillis();
   System.out.println("--------------------------------------------------------------");
   System.out.println("Ticks used :" + (stopTicks - startTicks));
   System.out.println("Errors detected :" + errorCount);
   System.exit(errorCount);
}



/**
     * Returns true if that key has a value
     * in the map. The key is considered to be a complete key.
     * @author Michael Winkler, doubleSlash GmbH
     *
     * @return boolean
     * @param partialKey java.lang.String The partial key to check for.
     */
public boolean containsKey(String fullKey) {
   synchronized (mutex_) {
     Set associations = fragments_.getAssociations(fullKey);
     return (associations != null) && (associations.size() > 0);
   }
}



   /**
      * Returns an array of all stored fragments that will lead to a subset of results.
      * @author Michael Winkler, doubleSlash GmbH
      *
      * @return java.lang.String[]
      */
   public String[] getFragmentsArray() {
    String[] returnValue = null;
    Object[] keysArray = fragments_.getKeysArray();
    returnValue = new String[keysArray.length];
    System.arraycopy(keysArray, 0, returnValue, 0, returnValue.length);
    return returnValue;
   }



   /**
      * This is the number of chars a single fragment has. It also limits the
      * amount of characters usable to put or get an element.
      * @author Michael Winkler, doubleSlash GmbH
      *
      * @return int
      */
   public int getFragmentSize() {
    return fragmentSize_;
   }



/**
   * Puts an element into that StringHashMap
   * @author Michael Winkler, doubleSlash GmbH
   *
   * @param key java.lang.String A single STring word, not containing any blanks.
   * @param value java.lang.Object
   */
public void pPut(String key, Object value) {

   // check params
   if (key.length() < fragmentSize_)
     throw new IllegalArgumentException("the key put into the FragmentMatcher MUST be of length >= fragmentSize!");

   // check for case sensitivity
   if (!caseSensitive_)
     key = key.toLowerCase();

   // put into pairs_ vector
   Object[] pair = { key, value };
   pairs_.add(pair);
   Integer pairIndex = new Integer(pairs_.size() - 1);

   // create fragments and put them into the fragment associator.
   Set fragmentSet = pConvertToFragmentVector(key);
   Object[] fragmentArray = fragmentSet != null ? fragmentSet.toArray() : null;
   for (int index = 0; index < fragmentSet.size(); index++)
     fragments_.associate(fragmentArray[index], pairIndex);

   // put the whole key into the inherited hashtable
   if (key.length() > fragmentSize_)
     fragments_.associate(key, pairIndex);

   elementsAdded_ = true;
}



/**
   * Puts an element into that StringHashMap
   * @author Michael Winkler, doubleSlash GmbH
   *
   * @param key java.lang.String A single String word, may contain blanks.
   * @param value java.lang.Object
   */
public void put(String key, Object value) {

   // check params
   if (key == null)
     throw new IllegalArgumentException("parameter 'key' is nulll.");
   if (value == null)
     throw new IllegalArgumentException("parameter 'value' is nulll.");

   // should we separate for single words?
   // since ignoreSpaces_== false by default, we first check for the changed value.
   if (ignoreSpaces_) {
     synchronized (mutex_) {
       if (key.length() < fragmentSize_)
         throw new IllegalArgumentException("parameter 'key' MUST be of length >= fragmentSize!");
       pPut(key, value);
     }

   } else {
     StringColumnizer columnizer = new StringColumnizer(key, " ");
     synchronized (mutex_) {
       while (columnizer.hasMoreTokens()) {
         String token = columnizer.nextToken().trim();
         if (token.length() < fragmentSize_)
            throw new IllegalArgumentException(
              "the single word '"
                + token
                + "' of the key '"
                + key
                + "' MUST ALSO be of length >= fragmentSize!");
         pPut(token, value);
       }
     }
   }
}



   /**
      * Code to perform when this object is garbage collected.
      *
      * Any exception thrown by a finalize method causes the finalization to
      * halt. But otherwise, it is ignored.
      */
   protected void finalize() throws Throwable {
    // Insert code to finalize the receiver here.
    // This implementation simply forwards the message to super.  You may replace or supplement this.
    super.finalize();
   }



/**
   * Returns a value, starting from that key.
   * This key may be a substring of the original
   * key, but must contain of at least three characters.
   * @author Michael Winkler, doubleSlash GmbH
   *
   * @return java.lang.Vector The list of values associated with that partial key.
   * @param partialKey java.lang.String A partial key,
   * containing of at least three letters.
   */
private List pGetPairs(String partialKey) {

   // init
   List returnValue = new Vector();

   // check for case sensitivity
   if (!caseSensitive_)
     partialKey = partialKey.toLowerCase();

   // find all fragments
   Set fragmentSet = pConvertToFragmentVector(partialKey);
   Object[] fragmentArray = fragmentSet.toArray(new String[0]);

   // generate initial cut

   Set cut = fragments_.getAssociations(fragmentArray[0]);
   boolean stopLoop = (cut == null) || (cut.size() < 1);
   int fragmentsCount = fragmentSet.size();
   if (fragmentsCount > 0) {
     for (int index = 0;(index < fragmentsCount) && !stopLoop; index++) {
       Set associations = fragments_.getAssociations(fragmentArray[index]);
       if ((associations != null) && (associations.size() > 0)) {
         // create cut with existing hits
         cut.retainAll(associations);
       } else {
         // huh, no hit for a fragment? CANCEL!
         cut.clear();
         break;
       }
     }
   }

   // did we find something?
   if ((cut != null) && (cut.size() > 0)) {
     List foundIndexes = new Vector(cut);
     CollectionHelper.removeDoubles(foundIndexes, true);

     // check validity of found results
     for (int index = 0; index < foundIndexes.size(); index++) {
       Integer foundIndex = (Integer) foundIndexes.get(index);

       // get the according pair
       Object[] pair = (Object[]) pairs_.get(foundIndex.intValue());

       // check wether the key string is contained.
       String key = pair[0].toString();
       if (key.indexOf(partialKey) > -1)
         returnValue.add(pair);
     }
   }

   return returnValue;
}



/**
      * Creates a new pattern matcher with the defined fragment size, which must be > 1!
      */
public FragmentMatcher(int fragmentSize, boolean ignoreSpaces) {
   this();
   if (fragmentSize < 2)
     throw new IllegalArgumentException("fragmentSize MUST be greater 1!");
   fragmentSize_ = fragmentSize;
   ignoreSpaces_ = ignoreSpaces;
}



/**
* Creates a new pattern matcher with a default fragment size.
* If ignoreSpaces if true, then spaces are treated as normal characters,
* and there is no word splitting done in <tt>FragmentMatcher.put(String, Object)</tt>.
* @param boolean ignoreSpaces If true, then spaces are treated as normal characters.
* @see FragmentMatcher.put(String,Object)
*/
public FragmentMatcher(boolean ignoreSpaces) {
   this();
   ignoreSpaces_ = ignoreSpaces;
}



   /**
   * Insert the method's description here.
   * Creation date: (10.11.2000 11:10:56)
   * @return boolean
   */
   public boolean isIgnoreSpaces() {
    return ignoreSpaces_;
   }



/**
    * Creates a new pattern matcher with the defined fragment size, which must be > 1!
    */
public FragmentMatcher(int fragmentSize, boolean ignoreSpaces, boolean caseSensitive) {
   this();
   if (fragmentSize < 2)
     throw new IllegalArgumentException("fragmentSize MUST be greater 1!");
   fragmentSize_ = fragmentSize;
   ignoreSpaces_ = ignoreSpaces;
   caseSensitive_ = caseSensitive;
}



/**
* Creates a new pattern matcher with a default fragment size.
* If ignoreSpaces if true, then spaces are treated as normal characters,
* and there is no word splitting done in <tt>FragmentMatcher.put(String, Object)</tt>.
* @param boolean ignoreSpaces If true, then spaces are treated as normal characters.
* @param boolean caseSensitive If true, then "anyText" allways will return the same results as "AnYtExT".
* @see FragmentMatcher.put(String,Object)
*/
public FragmentMatcher(boolean ignoreSpaces, boolean caseSensitive) {
   super();
   ignoreSpaces_ = ignoreSpaces;
   caseSensitive_ = caseSensitive;
}



/**
 * Returns wether or not this FragmentMatcher is case sensitive.
 * @author M.Winkler, doubleslash NetBusiness GmbH, Friedrichshafen
 * @return boolean
 */
public boolean isCaseSensitive() {
   return caseSensitive_;
}



/**
 * Sets a new case sensitivity. This method is only callable BEFORE any elements are put into it.
 * @author M.Winkler, doubleslash NetBusiness GmbH, Friedrichshafen
 * @param newCaseSensitive boolean
 */
public void setCaseSensitive(boolean newCaseSensitive) {
   if (elementsAdded_)
    throw new IllegalStateException("method not callable - there already have been added some elements.");
   caseSensitive_ = newCaseSensitive;
}



/**
     * Returns a value, starting from that key.
     * This key may be a substring of the original
     * key, but must contain of at least three characters.
     * @author Michael Winkler, doubleSlash GmbH
     *
     * @return java.lang.Vector The list of values associated with that partial key.
     * @param partialKey java.lang.String A partial key,
     * containing of at least three letters.
     */
public Set get(String partialKey) {

   // check params
   if (partialKey == null)
     throw new IllegalArgumentException("cannot handle null pointers.");
   Set returnValue = new HashSet();
   if (ignoreSpaces_) {
     returnValue = pGet(partialKey);
   } else {
     StringColumnizer columnizer = new StringColumnizer(partialKey, " ");
     synchronized (mutex_) {
       while (columnizer.hasMoreTokens()) {
         String token = columnizer.nextToken().trim();
         if (token.length() < fragmentSize_)
            throw new IllegalArgumentException("the single words of the key put into the FragmentMatcher MUST be of length >= fragmentSize!");
         Set partialResult = pGet(token);

         if ((partialResult != null) && (partialResult.size() > 0))
            returnValue.addAll(partialResult);
       }
     }
   }
   return returnValue;
}



/**
      * @author Michael Winkler, doubleSlash GmbH
      *
      * @return Vector
      * @param completeString java.lang.String
      */
private Set pConvertToFragmentVector(String completeString) {
   int fragmentCount = completeString.length() - fragmentSize_ + 1;
   Set returnValue = new TreeSet();
   for (int index = 0; index < fragmentCount; index++) {
     int endIndex = index + fragmentSize_;
     String subString = completeString.substring(index, endIndex);
     returnValue.add(subString);
   }
   return returnValue;
}



/**
     * Returns a value, starting from that key.
     * This key may be a substring of the original
     * key, but must contain of at least three characters.
     * @author Michael Winkler, doubleSlash GmbH
     *
     * @return Set The list of values associated with that partial key.
     * @param partialKey java.lang.String A partial key,
     * containing of at least three letters.
     */
protected Set pGet(String partialKey) {
   if (partialKey == null) {
     throw new IllegalArgumentException("partial key MUST be different from null.");
   }
   if (partialKey.length() < fragmentSize_) {
     throw new IllegalArgumentException("partial key must have more characters than a fragment has.");
   }
   List foundPairs = null;

   // call private pair getter method
   foundPairs = pGetPairs(partialKey);

   // Retrieve the values
   Set returnValue = new HashSet();
   if ((foundPairs != null) && (foundPairs.size() > 0)) {
     for (int index = 0; index < foundPairs.size(); index++) {
       Object[] pair = (Object[]) foundPairs.get(index);
       returnValue.add(pair[1]);
     }
   }
   return returnValue;
}
}